package nr.midi;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Vector;

import javax.sound.midi.MetaMessage;
import javax.sound.midi.MidiDevice;
import javax.sound.midi.MidiDevice.Info;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.Receiver;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.SysexMessage;
import javax.sound.midi.Transmitter;

import nr.music.Chord;
import nr.music.Tone;

import org.apache.log4j.Logger;

public class MidiReceiver implements Receiver {
	static private Logger log = Logger.getLogger(MidiReceiver.class);

	private MessageReceiver messageReceiver;
	private long toleranceMillis = 50l;

	private ArrayList<MidiReceiverListener> listeners = new ArrayList<MidiReceiverListener>();
	private MidiDevice device;

	/**
	 * @param receiver
	 */
	public void startReceivingMidi() {
		// Take the first device as default.
		if (device == null) {
			Vector<Info> infos = getDeviceList();
			if (infos.size() > 0) {
				Info info = infos.firstElement();
				setDevice(info);
			}
		}
	}

	public void stopReceivingMidi() {
		close();
	}

	public Vector<Info> getDeviceList() {
		Vector<Info> result = new Vector<Info>();
		Info[] infos = MidiSystem.getMidiDeviceInfo();
		for (Info info : infos) {
			//Only add available devices to the list.
			try {
				MidiDevice midiDevice = MidiSystem.getMidiDevice(info);
				if(midiDevice != null){
					midiDevice.getTransmitter();
				}
				result.add(info);
			} catch (MidiUnavailableException e) {
				log.warn("Device " + info + " not available.");
			}
		}
		return result;
	}

	public MidiDevice getDevice() {
		return device;
	}

	public void setDevice(Info deviceInfo) {
		try {
			if (device != null) {
				device.close();
			}
			device = MidiSystem.getMidiDevice(deviceInfo);
			log.info("device-Class:" + device.getClass().getName());
			if (device != null) {
				if (!device.isOpen()) {
					device.open();
				}
				if (messageReceiver == null) {
					messageReceiver = new MessageReceiver();
				}
				Transmitter transmitter = device.getTransmitter();
				transmitter.setReceiver(this);
			}
		} catch (MidiUnavailableException e) {
			log.error("device not available:" + deviceInfo, e);
			if (device != null && device.isOpen()) {
				device.close();
			}
			device = null;
		}
	}

	public void setDevice(MidiDevice device) {
		if (this.device != null && this.device.isOpen()) {
			this.device.close();
		}
		this.device = device;
	}

	public void send(MidiMessage message, long timeStamp) {
		// we only need ShortMessage
		if (message instanceof ShortMessage) {
			ShortMessage sm = (ShortMessage) message;
			Tone tone = decodeMessage(sm);
			if (tone != null && tone.getVelocity() != 0) {
				messageReceiver.newToneReceived(tone);
			}
		} else if (message instanceof SysexMessage) {
			log.debug("ignore SysexMessage");
		} else if (message instanceof MetaMessage) {
			log.debug("ignore MetaMessage");
		} else {
			log.debug("ignore unknown message type");
		}

	}

	private class MessageReceiver implements Runnable {
		private boolean isRunning = true;
		Chord chord = new Chord();

		public MessageReceiver() {
			Thread t = new Thread(this);
			t.start();
		}

		public void newToneReceived(Tone tone) {
			log.debug("newToneReceived");
			synchronized (chord) {
				log.debug("adding tone....");
				chord.addTone(tone);
				log.debug("tone added, notify Thread....");
				chord.notifyAll();
			}
		}

		public void run() {
			try {
				int status = 0;
				long startTime = 0;
				Chord chordCopy = null;
				while (isRunning) {
					switch (status) {
					case 0:
						synchronized (chord) {
							if (chord.getToneCount() > 0) {
								status = 1;
								startTime = System.currentTimeMillis();
								break;
							}
							try {
								chord.wait(10000l);
							} catch (InterruptedException e) {
								log.debug("InterruptedException....");
							}
						}
						break;
					case 1:
						synchronized (chord) {
							long wait = toleranceMillis
									- (System.currentTimeMillis() - startTime);
							if (wait > 0) {
								try {
									chord.wait(wait);
								} catch (InterruptedException e) {
									log.debug("InterruptedException....");
								}
							} else {
								log.debug("cloning chord:" + chord);
								chordCopy = chord.clone();
								chord.removeAll();
								chord.setKeyTone(null);
								status = 2;
							}
						}
						break;
					case 2:
						status = 0;
						if (chordCopy.getToneCount() == 1) {
							MidiReceiverEvent event = new MidiReceiverEvent(
									chordCopy.getKeyTone());
							notifyListeners(event);
						} else {
							MidiReceiverEvent event = new MidiReceiverEvent(
									chordCopy);
							notifyListeners(event);
						}

					}
				}
			} catch (Exception e) {
				log.error("MidiReceiver-Thread terminated!", e);
			}
		}

		public void stop() {
			isRunning = false;
			messageReceiver = null;
		}
	}

	public void close() {
		if (messageReceiver != null) {
			messageReceiver.stop();
		}
		if (device != null) {
			device.close();
		}
	}

	public Tone decodeMessage(ShortMessage message) {
		Tone result = null;
		// we only need noteOn and Off
		switch (message.getCommand()) {
		case 0x80: {
			// noteOff
			int keyNumber = message.getData1();
			result = new Tone(keyNumber);
			log.debug("created Tone from midi:" + result);
			break;

		}
		case 0x90:
			// noteOn
			int keyNumber = message.getData1();
			int velo = message.getData2();
			result = new Tone(keyNumber);
			result.setVelocity(velo);
			log.debug("created Tone from midi:" + result);
			break;
		}
		if (message.getCommand() != 0xF0) {
			// int nChannel = message.getChannel() + 1;
			// String strChannel = "channel " + nChannel + ": ";
		}
		return result;
	}

	public void addMidiReceiverListener(MidiReceiverListener listener) {
		listeners.add(listener);
	}

	public void removeMidiReceiverListener(MidiReceiverListener listener) {
		listeners.remove(listener);
	}

	private void notifyListeners(MidiReceiverEvent event) {
		log.debug("notifyListeners, event:" + event);
		Iterator<MidiReceiverListener> iter = listeners.iterator();
		while (iter.hasNext()) {
			MidiReceiverListener listener = iter.next();
			listener.midiDataReceived(event);
		}
	}

	public long getToleranceMillis() {
		return toleranceMillis;
	}

	public void setToleranceMillis(long toleranceMillis) {
		this.toleranceMillis = toleranceMillis;
	}

}
