package nr.music;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.Vector;

import org.apache.log4j.Logger;

public class Chord {
	static private Logger log = Logger.getLogger(Chord.class);
	
	//all tones of this chord
	private ArrayList<Tone> tones = new ArrayList<Tone>();
	//the key in which this chord is
	private Tone keyTone;
	//the key in which this cadence is
	private Tone cadenceKeyTone;
	//a symbol to show on screen
	private ChordSymbol symbol;
	//normally empty, when filled only these inversions are allowed
	private Vector<Integer> allowedInversions = new Vector<Integer>();
	
	public Chord() {
	}
	
	/**
	 * first note will be the determine key of the chord
	 * @param tones
	 */
	public Chord(Tone...tones) {
		for (int i = 0; i < tones.length; i++) {
			addTone(tones[i]);
		}
	}
	
	public void addTone(Tone tone) {
		if(tones.contains(tone)) return;
		if(keyTone == null) {
			setKeyTone(tone);
		}
		tones.add(tone);
		Collections.sort(tones);
	}
	
	public Collection<Tone> getTones() {
		return tones;
	}
	
	public int getToneCount() {
		return tones.size();
	}
	
	public void setKeyTone(Tone keyNote) {
		this.keyTone = keyNote;
		setRefsToSymbol();
	}

	public void setCadenceKeyTone(Tone cadenceKeyTone) {
		this.cadenceKeyTone = cadenceKeyTone;
		setRefsToSymbol();
	}

	private void setRefsToSymbol() {
		if (symbol != null ) {
			if(keyTone != null) {
				String keyName = keyTone.getName();
				symbol.setRefValue("key", keyName);
				if(keyName.length() == 2) {
					symbol.setRefValue("signlesskey", keyName.substring(0,1));
					symbol.setRefValue("sign", keyName.substring(1,2));
				}else {
					symbol.setRefValue("signlesskey", keyName);
					symbol.setRefValue("sign", "");
				}
			}
			if(cadenceKeyTone != null) {
				String cadenceKeyName = cadenceKeyTone.getName();
				symbol.setRefValue("cadencekey", cadenceKeyName);
				if(cadenceKeyName.length() == 2) {
					symbol.setRefValue("signlesscadencekey", cadenceKeyName.substring(0,1));
					symbol.setRefValue("cadencesign", cadenceKeyName.substring(1,2));
				}else {
					symbol.setRefValue("signlesscadencekey", cadenceKeyName);
					symbol.setRefValue("cadencesign", "");
				}
			}
		}
	}
	
	public Tone getKeyTone() {
		return keyTone;
	}
	
	public ChordSymbol getSymbol() {
		setRefsToSymbol();
		return symbol;
	}

	public void setSymbol(ChordSymbol symbol) {
		this.symbol = symbol;
	}

	public void removeTone(Tone note) {
		tones.remove(note);
	}
	
	public void removeAll() {
		tones.clear();
	}
	
	public void setAllowedInversions(Vector<Integer> allowedInversions) {
		this.allowedInversions = allowedInversions;
	}
	
	/**
	 * checks if this is an inversion of inversionChord and return the number (1 for first, 2 for second etc.)
	 * or -1 if it's not an inversion. If there are allowedInversions (@see #setAllowedInversions(Vector))
	 * and this is an inversion of inversionChord  and this inversion is not allowed the method returns -99.
	 * Important: the keyNote is ignored!
	 * 
	 * @param inversionChord
	 * @return the inversion, -1 if it's not an inversion or -99 if the inversion is not allowed
	 */
	public int isInversion(Chord inversionChord) {
		int ret = -1;
		//If the octave is ignored we also need to compare teh list size
		//to see if the same tones exist in different octaves. 
		if(tones.containsAll(inversionChord.tones) && inversionChord.tones.containsAll(tones) 
				&& inversionChord.tones.size() == tones.size()) {
			//base form
			ret = 0;
		}else {
			Tone baseTone = new Tone(tones.get(0));
			Chord clone = inversionChord.clone();
			int baseIdx = clone.tones.indexOf(baseTone);
			if(baseIdx != -1) {
				//make the base chord 
				int i = baseIdx - 1;
				for(;i >= 0;i--) {
					Tone tone = clone.tones.get(i);
					tone.setOctave(tone.getOctave() + 1);
				}
				//...and compare
				if(tones.containsAll(clone.tones) && clone.tones.containsAll(tones)) {
					ret = baseIdx;
				}
			}else {
				baseTone.setOctave(baseTone.getOctave() + 1);
				baseIdx = clone.tones.indexOf(baseTone);
				if(baseIdx != -1) {
					int i = baseIdx;
					int size = clone.getTones().size();
					for(;i < size;i++) {
						Tone tone = clone.tones.get(i);
						tone.setOctave(tone.getOctave() - 1);
					}
					if(tones.containsAll(clone.tones) && clone.tones.containsAll(tones)) {
						ret = baseIdx;
					}
				}
			}
		}
		if(ret != -1 && inversionChord.allowedInversions.size() > 0 && !inversionChord.allowedInversions.contains(new Integer(ret))) {
			ret = -99;
		}
		return ret;
	}

	public Chord getInversion(int inversion) {
		if(getToneCount() == 0) {
			return new Chord();
		}
		inversion = inversion % getToneCount();
		Chord inversionChord = clone();
		for(int i = 0; i < inversion; i++) {
			ArrayList<Tone> inversionTones = (ArrayList<Tone>) inversionChord.getTones();
			Tone base = inversionTones.remove(0);
			base.setOctave(base.getOctave() + 1);
			inversionChord.addTone(base);
		}
		log.debug("created inversion:" + inversionChord);
		return inversionChord;
	}
	
	@Override
	public String toString() {
		StringBuffer ret = new StringBuffer("Chord[KeyTone:");
		ret.append(getKeyTone()).append(", all Tones:");
		Iterator<Tone> iter = tones.iterator();
		StringBuffer temp = new StringBuffer("");
		while (iter.hasNext()) {
			Tone next = iter.next();
			temp.append(",").append(next);
		}
		if(temp.length()>0) {
			ret.append(temp.substring(1));
		}
		ret.append("]");
		return ret.toString();
	}
	
	@Override
	public boolean equals(Object obj) {
		if (obj != null && obj instanceof Chord) {
			Chord that = (Chord) obj;
			Iterator<Tone> iter = that.getTones().iterator();
			while(iter.hasNext()) {
				 if(!tones.contains(iter.next())){
					 return false;
				 }
			}
			if(!that.getKeyTone().equals(getKeyTone())) {
				return false;
			}
			return true;
		}
		return super.equals(obj);
	}
	
	public Chord clone() {
		Chord newChord = new Chord();
		Tone newKeyTone = new Tone(getKeyTone());
		newChord.setKeyTone(newKeyTone);
		if(tones.contains(getKeyTone())) {
			newChord.addTone(newKeyTone);
		}
		Iterator<Tone> iter = tones.iterator();
		while (iter.hasNext()) {
			Tone tone = iter.next();
			if(tone != getKeyTone()) {
				Tone newTone = new Tone(tone);
				newChord.addTone(newTone);
			}
		}
		ChordSymbol symbol = getSymbol();
		if(symbol != null) {
			newChord.setSymbol(new ChordSymbol(symbol));
		}
		if(allowedInversions.size() > 0) {
			Vector<Integer> invs = new Vector<Integer>();
			invs.addAll(allowedInversions);
			newChord.setAllowedInversions(invs);
		}
		return newChord;
	}

}
