package nr.music;

import java.util.Arrays;
import java.util.List;


public class Tone implements Comparable<Tone>{
	
	public enum Sign {SHARP, FLAT};
	private Sign sign;
	
	//The order is important for calculating the absolutHeight!
	public static final String[] keyNamesSharp = { "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B" };
	public static final String[] keyNamesFlat = { "C", "Db", "D", "Eb", "E", "F", "Gb", "G", "Ab", "A", "Bb", "B" };

	
	private String name;
	private int velocity = 0;
	private int octave = 0;
	private int absolutHeight = -1;

//	static private Category log = Category.getInstance(Tone.class);
	
	public Tone(int absolutHeight) {
		setAbsolutHeight(absolutHeight);
	}
	public Tone(String name, int octave) {
		this(name, 127, octave);
	}
	public Tone(String name, int velocity, int octave) {
		setName(name);
		setVelocity(velocity);
		setOctave(octave);
	}
	
	public Tone(Tone tone) {
		this(tone.getName(), tone.getVelocity(), tone.getOctave());
		setSign(tone.getSign());
	}

	public Tone(String name) {
		this(name, 127, 3);
	}
	
	public String getName() {
		return name;
	}

	public void setName(String name) {
		List<String> keyNamesFlatList = Arrays.asList(keyNamesFlat);
		List<String> keyNamesSharpList = Arrays.asList(keyNamesSharp);
		if(keyNamesFlatList.contains(name)) {
			this.name = name;
			if(name.length() > 1) {
				sign = Sign.FLAT;
			}
		}else if(keyNamesSharpList.contains(name)) {
			this.name = name;
			if(name.length() > 1) {
				sign = Sign.SHARP;
			}
		}else {
			throw new IllegalArgumentException("name must be one of this:" + keyNamesSharp + keyNamesFlat);
		}
		recalcAbsolutHeight();
	}

	public int getVelocity() {
		return velocity;
	}

	public void setVelocity(int velocity) {
		this.velocity = velocity;
	}
	
	public int getOctave() {
		return octave;
	}
	public void setOctave(int octave) {
		this.octave = octave;
		recalcAbsolutHeight();
	}
	
	public Sign getSign() {
		return sign;
	}
	public void setSign(Sign sign) {
		this.sign = sign;
		setAbsolutHeight(absolutHeight);
	}

	/**
	 * @param obj
	 * @param absolutHeight
	 * @return
	 */
	public int recalcAbsolutHeight() {
		List<String> keyNamesList = null;
		if(sign == Sign.SHARP) {
			keyNamesList = Arrays.asList(keyNamesSharp);
		}else {
			keyNamesList = Arrays.asList(keyNamesFlat);
		}
		absolutHeight = keyNamesList.indexOf(getName()) + (getOctave() * 12);
		return absolutHeight;
	}
	
	public int getAbsolutHeight() {
		return recalcAbsolutHeight();
	}
	
	/**
	 * @param absoluteHeight (>-1, <128)
	 */
	public void setAbsolutHeight(int absoluteHeight) {
		assert absoluteHeight < 128 && absoluteHeight > -1:"absolutHeight >-1, <128";
		this.absolutHeight = absoluteHeight;
		int nNote = absoluteHeight % 12;
		if(sign == Sign.SHARP) {
			name = keyNamesSharp[nNote]; 
		}else {
			name = keyNamesFlat[nNote]; 
		}
		octave = absoluteHeight / 12;
	}
	
	@Override
	public boolean equals(Object obj) {
		if (obj != null && obj instanceof Tone) {
			Tone that = (Tone) obj;
//			log.debug("that.recalcAbsolutHeight():" +that.recalcAbsolutHeight() + " recalcAbsolutHeight():" + recalcAbsolutHeight());
			return that.recalcAbsolutHeight() == recalcAbsolutHeight();
		}
		return false;
	}

	public int compareTo(Tone obj) {
		if(obj == null) throw new NullPointerException("can't compare to null!");
		if (obj instanceof Tone) {
			Tone that = (Tone) obj;
			int thisTonHoehe = recalcAbsolutHeight();
			int thatTonHoehe = that.recalcAbsolutHeight();
			return thisTonHoehe - thatTonHoehe;
		}
		return 99999;
	}

	@Override
	public String toString() {
		return "Tone[name:"+ getName() + ", velocity:" + getVelocity() + ", octave:" + getOctave() + ", absolutHeight:" + recalcAbsolutHeight() + "]";
	}
	@Override
	public int hashCode() {
		return recalcAbsolutHeight();
	}
}
