package nr.music.chordprovider;

import java.util.ArrayList;

import nr.music.Tone;
import nr.music.Tone.Sign;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;

public class KeyProvider {
	private static final Logger log = Logger.getLogger(KeyProvider.class);

	public enum KeySequence {
		CHROMATIC, QUINTCIRCLE, QUARTCIRCLE, RANDOM
	};

	private KeySequence keySequence = KeySequence.QUINTCIRCLE;
	
	public enum SignMode {
		SIGN_FLAT {
			Sign getSign() {
				return Sign.FLAT;
			}
		},
		SIGN_SHARP {
			Sign getSign() {
				return Sign.SHARP;
			}
		},
		SIGN_RANDOM {
			Sign getSign() {
				Sign randomSign = null;
				if(Math.random() >= 0.5) {
					randomSign = Sign.FLAT;
				} else {
					randomSign = Sign.SHARP;
				}
				return randomSign;
			}
		};
		abstract Sign getSign();
	};
	
	private SignMode signMode = SignMode.SIGN_RANDOM;
	private Tone firstKey = new Tone(0);
	private ArrayList<Tone> keysToProvide = new ArrayList<Tone>();

	private ArrayList<Tone> disabledKeys = new ArrayList<Tone>();

	public KeyProvider() {
		SignMode.SIGN_FLAT.getSign();
		reset();
	}

	public static void main(String[] args) {
		BasicConfigurator.configure();
		KeyProvider provider = new KeyProvider();
//		provider.setKeySequence(KeySequence.RANDOM);
		provider.addDisabledKey(new Tone(6));
		KeySequence[] seq = {KeySequence.CHROMATIC, KeySequence.QUARTCIRCLE, KeySequence.QUARTCIRCLE, KeySequence.RANDOM}; 
		int callReset = 11;
		int callResort = 7;
		for (int i = 0; i < 100; i++) {
			log.debug("provider.next():" + provider.next());
			if(i % callReset == 0) {
				provider.reset();
			}
			if(i % callResort == 0) {
				provider.setKeySequence(seq[(int)(Math.random() * 4)]);
			}
		}

	}

	public void addDisabledKey(Tone key) {
		if (!disabledKeys.contains(key)) {
			Tone disKey = new Tone(key);
			disKey.setOctave(0);
			log.debug("adding disabledKey:" + disKey);
			disabledKeys.add(disKey);
			int idx;
			if((idx = keysToProvide.indexOf(disKey)) > -1) {
				keysToProvide.remove(idx);
			}
		}
	}

	public void removeDisabledKey(Tone key) {
		disabledKeys.remove(key);
	}

	public void removeAllDisabledKeys() {
		disabledKeys.clear();
	}

	public void reset() {
		log.debug("reset...");
		keysToProvide.clear();
		for (int i = 0; i < 12; i++) {
			Tone tone = new Tone(i);
			if (!disabledKeys.contains(tone)) {
				keysToProvide.add(tone);
			}
		}
		resort();
		log.debug("...reset");
	}

	private void resort() {
		log.debug("resort... KeySequence:" + keySequence);
		if (keysToProvide.size() != 0) {
			ArrayList<Tone> oldKeysToProvide = new ArrayList<Tone>(keysToProvide);
			keysToProvide.clear();
			if (keySequence == KeySequence.RANDOM) {
				int tempListSize;
				while ((tempListSize = oldKeysToProvide.size()) > 0) {
					int next = (int) (Math.random() * tempListSize);
					Tone nextKey = oldKeysToProvide.remove(next);
					log.debug("adding key to provide:" + nextKey);
					keysToProvide.add(nextKey);
				}
			} else {
				int halfToneDelta = 0;
				switch (keySequence) {
				case CHROMATIC:
					halfToneDelta = 1;
					break;
				case QUARTCIRCLE:
					halfToneDelta = 5;
					break;
				case QUINTCIRCLE:
					halfToneDelta = 7;
					break;
				case RANDOM:
					//ignore
					break;
				}
				Tone nextKey = new Tone(firstKey);
				for (int i = 0; i < 12; i++) {
					int idx;
					if ((idx = oldKeysToProvide.indexOf(nextKey)) > -1) {
						Tone tone = oldKeysToProvide.remove(idx);
						log.debug("adding key to provide:" + tone);
						keysToProvide.add(tone);
					}
					nextKey.setAbsolutHeight((nextKey.getAbsolutHeight() + halfToneDelta) % 12);
				}
			}
		}
		log.debug("...resort");
	}

	public Tone next() {
		Tone nextKey = null;
		if (keysToProvide.size() > 0) {
			nextKey = keysToProvide.remove(0);
			nextKey.setSign(signMode.getSign());
		}
		log.debug("nextKey:" + nextKey);
		return nextKey;
	}

	public KeySequence getKeySequence() {
		return keySequence;
	}

	public void setKeySequence(KeySequence keySequence) {
		this.keySequence = keySequence;
		resort();
	}

	public SignMode getSignMode() {
		return signMode;
	}

	public void setSignMode(SignMode signMode) {
		this.signMode = signMode;
	}

	public Tone getFirstKey() {
		return firstKey;
	}

	public void setFirstKey(Tone firstKey) {
		this.firstKey = firstKey;
	}

}
