package nr.music.xml;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Vector;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import nr.music.Cadence;
import nr.music.Chord;
import nr.music.ChordDefinition;
import nr.music.ChordSymbol;
import nr.music.ChordSymbol.Line;
import nr.music.Interval;
import nr.music.Tone;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

public class ChordXmlParser {
	static private Logger log = Logger.getLogger(ChordXmlParser.class);
	
	private Vector<Chord> chords = new Vector<Chord>();
	private Hashtable<String, ChordDefinition> chordDefinitions = new Hashtable<String, ChordDefinition>();
	private Vector<Cadence> cadences = new Vector<Cadence>();
	private HashMap<String, ChordSymbol> symbols = new HashMap<String, ChordSymbol>();
	
	File xmlFile;

	
	public ChordXmlParser() {
	}
	
	public ChordXmlParser(File xmlFile) {
		this.xmlFile = xmlFile;
	}
	
	public Vector<Chord> parse() {
		try {
			DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
			boolean exists = xmlFile.exists();
			log.debug("parsing file " + xmlFile + " exists:" + exists);
			if(exists) {
				chords.clear();
				Document document = builder.parse(xmlFile);
				NodeList nlist = document.getChildNodes();
				for(int idx=0; idx<nlist.getLength(); idx++) {
					Node node = nlist.item(idx);
					String nodeName = node.getNodeName();
					
					if (nodeName.equals("chords")) {
						parseChords(node);
					}
				}
				log.debug("parsed " + xmlFile + ": chords:" + chords.size() + ", chorddefinitions:" + chordDefinitions.size() + " symbols:" + symbols.size());
				return chords;
			}
			throw new FileNotFoundException("can't find file " + xmlFile);
		} catch (ParserConfigurationException e) {
			log.error("", e);
		} catch (FileNotFoundException e) {
			log.error("", e);
		} catch (SAXException e) {
			log.error("", e);
		} catch (IOException e) {
			log.error("", e);
		}
		return null;
	}
	
	private void parseChords(Node node) {
		NodeList nlist = node.getChildNodes();
		for(int idx=0; idx<nlist.getLength(); idx++) {
			Node child = nlist.item(idx);
			String nodeName = child.getNodeName();
			if (nodeName.equals("chordsymbol")) {
				ChordSymbol newSymbol = parseSymbol(child);
				if(newSymbol != null) {
					symbols.put(newSymbol.getName(), newSymbol);
					log.debug("chordsymbol added:" + newSymbol);
				}
			}else if (nodeName.equals("chorddefinition")) {
				ChordDefinition def = parseChordDefinition(child);
				chordDefinitions.put(def.getName(), def);
			}else if (nodeName.equals("cadence")) {
				Cadence def = parseCadence(child);
				cadences.add(def);
			}else if (nodeName.equals("chord")) {
				Chord newChord = parseChord(child);
				chords.add(newChord);
				log.debug("chord added:" + newChord);
			}
		}

	}

	private Cadence parseCadence(Node node) {
		log.debug("parsing Cadence...");
		Cadence cadence = new Cadence();
		NodeList nlist = node.getChildNodes();
		for(int idx=0; idx<nlist.getLength(); idx++) {
			Node child = nlist.item(idx);
			if(child.getNodeName().equals("chorddefinitionref")) {
				Node refAttrib = child.getAttributes().getNamedItem("ref");
				Node stufeAttrib = child.getAttributes().getNamedItem("stufe");
				if(refAttrib == null || stufeAttrib == null) {
					log.error("for cadences i need 2 attributes: ref, stufe");
				}else {
					String stufe = stufeAttrib.getNodeValue();
					Interval interval = new Interval(stufe);
					String refName = refAttrib.getNodeValue();
					ChordDefinition definition = chordDefinitions.get(refName);
					if(definition != null) {
						cadence.addChordDefinition(definition, interval);
						log.debug("added ChordDefinition:" + definition + ", interval:" + interval);
					}else {
						log.error("can't find chorddefinition with the name " + refName);
					}
					NodeList childNodes = child.getChildNodes();
					if(childNodes.getLength() > 0) {
						Vector<Integer> inversions = new Vector<Integer>();
						for(int i=0; i<childNodes.getLength(); i++) {
							Node inversionNode = childNodes.item(i);
							if(inversionNode.getNodeName().equals("inversion")) {
								Node noAttrib = inversionNode.getAttributes().getNamedItem("no");
								if(noAttrib == null) {
									log.error("for inversion i need attribute: no");
								}else {
									inversions.add(new Integer(noAttrib.getNodeValue()));
								}
							}
						}
						if(inversions.size() > 0) {
							definition.setAllowedInversions(inversions);
							log.debug("added " + inversions.size() + " inversions");
						}
					}
				}
			}
		}
		return cadence;
	}

	private ChordDefinition parseChordDefinition(Node node) {
		log.debug("parsing ChordDefinitions...");
		ChordDefinition definition = null;
		Node symbolAttrib = node.getAttributes().getNamedItem("symbol");
		Node nameAttrib = node.getAttributes().getNamedItem("name");
		if(symbolAttrib == null || nameAttrib == null) {
			log.error("for chorddefinition i need 2 attributes: symbol, name");
		}else {
			String name = nameAttrib.getNodeValue();
			String symbolName = symbolAttrib.getNodeValue();

			ChordSymbol symbol = new ChordSymbol(symbols.get(symbolName));
			definition = new ChordDefinition(name, symbol);
			NodeList childNodes = node.getChildNodes();
			Vector<Integer> allowedInversions = new Vector<Integer>();
			for (int i = 0; i < childNodes.getLength(); i++) {
				Node child = childNodes.item(i);
				if(child.getNodeName().equals("interval")) {
					Node nameAtt = child.getAttributes().getNamedItem("name");
					Interval interval = new Interval(nameAtt.getNodeValue());
//					if (interval == 0) {
//						log.error("Error: interval zero is not allowed!!");
//					}
					definition.addInterval(interval);
				}else if(child.getNodeName().equals("inversion")) {
					Node nameAtt = child.getAttributes().getNamedItem("no");
					if(nameAtt != null) {
						allowedInversions.add(new Integer(nameAtt.getNodeValue()));
					}
				}
			}
			definition.setAllowedInversions(allowedInversions);
		}
		return definition;
	}

	private ChordSymbol parseSymbol(Node node) {
		log.debug("parsing symbol...");
		ChordSymbol symbol = null;
		Node nameAttrib = node.getAttributes().getNamedItem("name");
		if(nameAttrib != null) {
			String nodeValue = nameAttrib.getNodeValue();
			log.debug("symbolName:" + nodeValue);
			symbol = new ChordSymbol(nodeValue);
			
			NodeList childNodes = node.getChildNodes();
			log.debug("childNodes.getLength():" + childNodes.getLength());
			for(int idx=0; idx<childNodes.getLength(); idx++) {
				Node child = childNodes.item(idx);
				Line line = null;
				String nodeName = child.getNodeName();
				log.debug("child.getNodeName():" + nodeName);
				if(nodeName.equals("baseline")) {
					log.debug("got baseline entry");
					line = Line.BASE;
				}else if (nodeName.equals("upperline")) {
					log.debug("got upperline entry");
					line = Line.UPPER;
				}else {
					log.debug("unknown node:" + nodeName);
					
				}
				NamedNodeMap attributes = child.getAttributes();
				if(attributes != null) {
					parseSymbolAttributes(attributes, symbol, line);
				}
			}
			
		}else {
			log.debug("no symbol name - ignoring symbol");
		}
		return symbol;
	}

	/**
	 * @param symbol
	 * @param child
	 * @param line
	 */
	private void parseSymbolAttributes(NamedNodeMap attributes, ChordSymbol symbol, Line line) {
		log.debug("parsing SymbolAttributes");
		Node textAttrib = attributes.getNamedItem("text");
		Node textRefAttrib = attributes.getNamedItem("textref");
		Node sizeAttrib = attributes.getNamedItem("size");
		boolean small = false;
		if((sizeAttrib != null && sizeAttrib.getNodeValue().equals("small"))
				|| line == Line.UPPER) {
			log.debug("small symbol");
			small = true;
		}
		if(textAttrib != null) {
			String symbolText = textAttrib.getNodeValue();
			log.debug("symbolText:" + symbolText);
			symbol.addText(line, symbolText, small);
		}
		if(textRefAttrib != null) {
			String refName = textRefAttrib.getNodeValue();
			log.debug("refName:" + refName);
			symbol.addRef(line, refName, small);
		}
	}

	private Chord parseChord(Node node) {
		log.debug("parsing chord...");
		Chord chord = new Chord();
		Node displayAttrib = node.getAttributes().getNamedItem("display");
		if(displayAttrib != null) {
			String displayText = displayAttrib.getNodeValue();
			ChordSymbol symbol = new ChordSymbol(displayText);
			log.debug("found display attrib:" + displayText);
			symbol.addText(Line.BASE, displayText, false);
			chord.setSymbol(symbol);
		}
		Node octaveAttrib = node.getAttributes().getNamedItem("octave");
		int octave = -1;
		if(octaveAttrib != null) {
			octave = new Integer(octaveAttrib.getNodeValue()).intValue();
		}
		Node symbolAttrib = node.getAttributes().getNamedItem("symbol");
		if(symbolAttrib != null) {
			String symbolName = symbolAttrib.getNodeValue();
			ChordSymbol chordSymbol = new ChordSymbol(symbols.get(symbolName));
			chord.setSymbol(chordSymbol);
		}
		NodeList nlist = node.getChildNodes();
		for(int idx=0; idx<nlist.getLength(); idx++) {
			Node childNode = nlist.item(idx);
			String nodeName = childNode.getNodeName();
			if (nodeName.equals("note")) {
				Tone note = null;
				if(octave == -1) {
					note = parseTone(childNode);
				}else {
					note = parseTone(childNode, octave);
				}
				chord.addTone(note);
			}
		}
		return chord;
	}

	private Tone parseTone(Node childNode) {
		return parseTone(childNode, -1);
	}
	
	private Tone parseTone(Node childNode, int defaultOctave) {
		Tone note = null;
		Node attrib = childNode.getAttributes().getNamedItem("name");
		if(attrib != null) {
			String value = attrib.getNodeValue();
			note = new Tone(value);
		}else {
			log.error("you must give a note-name for a note");
		}
		attrib = childNode.getAttributes().getNamedItem("octave");
		if(attrib != null) {
			note.setOctave(new Integer(attrib.getNodeValue()).intValue());
		}else {
			note.setOctave(defaultOctave);
		}

		return note;
	}

	public Vector<Chord> getChords() {
		return chords;
	}
	
	public Hashtable<String, ChordDefinition> getChordDefinitions() {
		return chordDefinitions;
	}

	public File getXmlFile() {
		return xmlFile;
	}

	public void setXmlFile(File xmlFile) {
		this.xmlFile = xmlFile;
	}
	
	public Vector<Cadence> getCadences() {
		return cadences;
	}

	
}
